#include <windows.h>
#include <stdlib.h>
#include <mmreg.h>
#include <math.h>
#include "labs.h"

#pragma option -w-sig

HANDLE hInstance;
HWND hWndApp;

const DWORD dwDataSize=88200;

class SOUND
 {
  WAVEFORMATEX Format;
  HWAVEOUT hWaveOut;
  HANDLE hData; // handle of waveform data memory
  HGLOBAL hWaveHdr;
  LPWAVEHDR lpWaveHdr;
  HPSTR lpData; // pointer to waveform data memory
 public:
  bool Started;
  SOUND();
  bool Init(WORD nChannels, WORD wBitsPerSample, DWORD nSamplesPerSec);
  ~SOUND();
  void SetParam(char form, WORD freq);
  void Start();
  void Stop();
  void pSyn(WORD freq);
  void pTriangle(WORD freq);
  void pSquare(WORD freq);
 };

SOUND snd;

void ErrorBox(char *sz)
 {
  MessageBox(hWndApp, sz, NULL, MB_OK | MB_ICONEXCLAMATION);
 }

MMRESULT MMError(MMRESULT r)
 { char text[256];
   if (r)
    {
      waveOutGetErrorText(r, text, sizeof(text));
      ErrorBox(text);
    }
   return r;
 }

SOUND::SOUND()
 {
   hWaveOut = 0;
   hData  = NULL;
   hWaveHdr = NULL;
   Started = false;
 }

SOUND::~SOUND()
 {
   Stop();
   if (hData) {GlobalUnlock(hData); GlobalFree(hData);}
   if (hWaveHdr) {GlobalUnlock(hWaveHdr); GlobalFree(hWaveHdr);}
   hData = NULL; hWaveHdr = NULL;
 }

bool SOUND::Init(WORD nChannels, WORD wBitsPerSample, DWORD nSamplesPerSec)
 {
    if (hData) this->~SOUND();
    Format.wFormatTag = WAVE_FORMAT_PCM;
    Format.nChannels = nChannels;
    Format.wBitsPerSample = wBitsPerSample;
    Format.nSamplesPerSec = nSamplesPerSec;
    Format.nBlockAlign = (Format.nChannels * Format.wBitsPerSample) >> 3;
    Format.nAvgBytesPerSec = Format.nSamplesPerSec * Format.nBlockAlign;
    Format.cbSize = 0;

    /* Allocate and lock memory for the waveform data. The memory
     * for waveform data must be globally allocated with
     * GMEM_MOVEABLE and GMEM_SHARE flags.  */
    if ((hData = GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE, dwDataSize))==NULL)
      { ErrorBox("Out of memory."); return false; }

    if ((lpData = (char*)GlobalLock(hData)) == NULL)
      {
        GlobalFree(hData); hData = NULL;
        ErrorBox("Failed to lock memory for data chunk.");
        return false;
      }

    /*
     * Allocate and lock memory for the header. This memory must
     * also be globally allocated with GMEM_MOVEABLE and
     * GMEM_SHARE flags.
     */
    hWaveHdr = GlobalAlloc(GMEM_MOVEABLE | GMEM_SHARE, (DWORD) sizeof(WAVEHDR));
    if (hWaveHdr == NULL)
      {
        GlobalUnlock(hData); GlobalFree(hData); hData = NULL;
        ErrorBox("Not enough memory for header.");
        return false;
      }

    lpWaveHdr = (LPWAVEHDR) GlobalLock(hWaveHdr);
    if (lpWaveHdr == NULL)
     {
       GlobalUnlock(hData); GlobalFree(hData); hData = NULL;
       GlobalUnlock(hWaveHdr); GlobalFree(hWaveHdr); hWaveHdr = NULL;
       ErrorBox("Failed to lock memory for header.");
       return false;
     }
    return true;
 }

void SOUND::Start()
 {
   if (!hData) return;
   if (Started) Stop();

   if (MMError(waveOutOpen((LPHWAVEOUT)&hWaveOut, WAVE_MAPPER,
             &Format, 0, 0, CALLBACK_NULL))) return;

   lpWaveHdr->lpData = lpData;
   lpWaveHdr->dwBufferLength = dwDataSize;
   lpWaveHdr->dwFlags = 0;
   lpWaveHdr->dwLoops = 500000;
   MMError(waveOutPrepareHeader(hWaveOut, lpWaveHdr, sizeof(WAVEHDR)));
   lpWaveHdr->dwFlags |= WHDR_BEGINLOOP | WHDR_ENDLOOP;

   if (MMError(waveOutWrite(hWaveOut, lpWaveHdr, sizeof(WAVEHDR))))
    {
      waveOutUnprepareHeader(hWaveOut, lpWaveHdr, sizeof(WAVEHDR));
      return;
    }
   Started = true;
 }

void SOUND::Stop()
 {
  if (Started)
   if (!MMError(waveOutReset(hWaveOut)))
    if (!MMError(waveOutUnprepareHeader(hWaveOut, lpWaveHdr, sizeof(WAVEHDR))))
     if (!MMError(waveOutClose(hWaveOut))) Started = false;
 }

void SOUND::SetParam(char form, WORD freq)
 {
  if (!hData) return;
  if (freq==0) {FillMemory(lpData, dwDataSize, 0); return;}
  switch (form)
   {
    case FORM_SYN: pSyn(freq); break;
    case FORM_TRIANGLE: pTriangle(freq); break;
    case FORM_SQUARE: pSquare(freq); break;
   }
 }

void SOUND::pSyn(WORD freq)
 {
    double x=freq*(PI*2.0)/Format.nSamplesPerSec;
    short *lptr = (short*)lpData;
    for (DWORD i=0; i<dwDataSize/2; i++, lptr++) *lptr = short(sin(i*x)*32765);
 }

void SOUND::pTriangle(WORD freq)
 {
  double y, p;
  double T = Format.nSamplesPerSec/(double)freq;
  short *lptr = (short*)lpData;
  for (DWORD i=0; i<dwDataSize/2; i++, lptr++)
   {
    p = modf(i/T+0.25, &y);
    *lptr =(p<0.5) ? p*4*32765-32765 : 3*32765-4*32765*p;
   }
 }

void SOUND::pSquare(WORD freq)
 {
  double T = Format.nSamplesPerSec/(double)freq;
  double y;
  short *lptr = (short*)lpData;
  for (DWORD i=0; i<dwDataSize/2; i++, lptr++)
    *lptr = (modf(i/T, &y) < 0.5) ? 32765 : -32765;
 }

#define IB 19
#define IE 200
#define LGB 1.3010299956639812
#define LGE 4.3010299956639812
#define k ((IE-IB)/(LGE-LGB))
#define b (IB-k*LGB)

WORD ifx(WORD x)
 {
  if (x<=IB) return x;
  return log(x)*k/LN10+b+0.5;
 }

WORD xfi(WORD i)
 {
  if (i<=IB) return i;
  return exp((i-b)*LN10/k);
 }

#undef b
#undef k
#undef LGE
#undef LGB
#undef IB

WORD GetFreq(HWND hwndDlg)
 {
  BOOL flag;
  DWORD w = GetDlgItemInt(hwndDlg, IDC_EDIT, &flag, false);
  if (!flag) w = 0; else if (w>20000) {w = 20000; flag=FALSE; }
  if (!flag) SetDlgItemInt(hwndDlg, IDC_EDIT, w, false);
  return w;
 }

#pragma argsused
BOOL CALLBACK AboutProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM  lParam)
 {
  switch (uMsg)
   {
    case WM_INITDIALOG: return TRUE;
    case WM_COMMAND:
     if (LOWORD(wParam)==IDOK) {EndDialog(hwndDlg, 0); return TRUE;}
    default: return FALSE;
   }
 }

#pragma argsused
BOOL CALLBACK MainProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM  lParam)
 {
  static char form;
  static SCROLLINFO si = {{sizeof(SCROLLINFO)}, {0}, {0}, {IE}, {10}};
  WORD freq;

  switch (uMsg)
   {
    case WM_INITDIALOG:
      SendMessage(hwndDlg, WM_SETICON, ICON_SMALL,
        (LONG)LoadIcon(hInstance, MAKEINTRESOURCE(IDI_MAINICON)));
      SendDlgItemMessage(hwndDlg, IDC_SYN, BM_SETIMAGE, IMAGE_ICON,
        (LPARAM)LoadImage(hInstance, MAKEINTRESOURCE(IDI_SYN), IMAGE_ICON,
         0, 0, LR_DEFAULTCOLOR));
      SendDlgItemMessage(hwndDlg, IDC_TRIANGLE, BM_SETIMAGE, IMAGE_ICON,
        (LPARAM)LoadImage(hInstance, MAKEINTRESOURCE(IDI_TRIANGLE), IMAGE_ICON,
         0, 0, LR_DEFAULTCOLOR));
      SendDlgItemMessage(hwndDlg, IDC_SQUARE, BM_SETIMAGE, IMAGE_ICON,
        (LPARAM)LoadImage(hInstance, MAKEINTRESOURCE(IDI_SQUARE), IMAGE_ICON,
         0, 0, LR_DEFAULTCOLOR));
      CheckRadioButton(hwndDlg, IDC_SYN, IDC_SQUARE, IDC_SYN); form = FORM_SYN;
      freq=GetFreq(hwndDlg);
      si.nPos=ifx(freq);
      si.fMask = SIF_POS | SIF_RANGE;
      SendDlgItemMessage(hwndDlg, IDC_SCROLLBAR, SBM_SETSCROLLINFO,
         TRUE, (LPARAM) &si);
      snd.SetParam(form, freq);
      break;

    case WM_CLOSE:
      DestroyWindow(hwndDlg);
      PostQuitMessage(0);
      break;

    case WM_HSCROLL:
      switch LOWORD(wParam)
       {
        case SB_THUMBTRACK: si.nPos = HIWORD(wParam); break;
        case SB_LINELEFT: if (si.nPos>0) si.nPos--; break;
        case SB_LINERIGHT: if (si.nPos<IE) si.nPos++; break;
        case SB_PAGELEFT: si.nPos>10 ? si.nPos-=10 : si.nPos=0; break;
        case SB_PAGERIGHT: si.nPos<IE-10 ? si.nPos+=10 : si.nPos=IE; break;
        default: return FALSE;
       }
      freq = xfi(si.nPos);
      SetDlgItemInt(hwndDlg, IDC_EDIT, freq, false);
      si.fMask = SIF_POS;
      SendDlgItemMessage(hwndDlg, IDC_SCROLLBAR, SBM_SETSCROLLINFO,
         TRUE, (LPARAM) &si);
      snd.SetParam(form, freq);
      break;

    case WM_COMMAND:
     switch (LOWORD(wParam))
      {
       case IDC_SYN:
       case IDC_TRIANGLE:
       case IDC_SQUARE:
          form = LOWORD(wParam)-IDC_SYN+FORM_SYN;
          snd.SetParam(form, GetFreq(hwndDlg));
          break;

       case IDC_ABOUT:
          DialogBox(hInstance, MAKEINTRESOURCE(IDD_ABOUTDIALOG),
             hwndDlg, (DLGPROC)AboutProc);
          break;

       case IDC_START:
          freq = GetFreq(hwndDlg);
          si.nPos = ifx(freq);
          si.fMask = SIF_POS;
          SendDlgItemMessage(hwndDlg, IDC_SCROLLBAR, SBM_SETSCROLLINFO,
             TRUE, (LPARAM) &si);
          snd.SetParam(form, freq);
          snd.Started ? snd.Stop() : snd.Start();
          SendDlgItemMessage(hwndDlg, IDC_START,
             WM_SETTEXT,0,(LPARAM)((snd.Started)?".":"."));
          break;

       case IDC_EDIT:
          if (HIWORD(wParam)==EN_KILLFOCUS)
           {
            freq = GetFreq(hwndDlg);
            si.nPos = ifx(freq);
            si.fMask = SIF_POS;
            SendDlgItemMessage(hwndDlg, IDC_SCROLLBAR, SBM_SETSCROLLINFO,
               TRUE, (LPARAM) &si);
            snd.SetParam(form, freq);
           }
          break;

       default: return FALSE;
      } break;

    default: return FALSE;
   }
  return TRUE;
 }

#pragma argsused
int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInst,
                   LPSTR lpszCmdLine, int nCmdShow)
{
 hInstance=hInst;
 static MSG lpMsg;

 if (!snd.Init(1, 16, 44100)) return 1;

 hWndApp=CreateDialogParam(hInstance, MAKEINTRESOURCE(IDD_MAINWINDOW),
      (HWND)NULL, (DLGPROC)MainProc, 0);
 if (!hWndApp) return 1;


 while (GetMessage(&lpMsg, NULL, 0, 0))
   if (!IsDialogMessage(hWndApp,&lpMsg))
   {
    TranslateMessage(&lpMsg);
    DispatchMessage(&lpMsg);
   }

 return lpMsg.wParam;
}


